"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional, Union
import asyncio
import sys
from concurrent.futures import ThreadPoolExecutor

from dbus_fast.aio import MessageBus
from dbus_fast import BusType, Message, MessageType
import dbus_fast
from bcc import __version__ as bcc_version
from packaging import version

from proton.vpn.daemon.split_tunneling import dbus_translator as translator
from proton.vpn.core.settings import SplitTunnelingConfig
from proton.vpn.split_tunneling import exceptions
from proton.vpn.split_tunneling import SplitTunneling

is_python_3_11_or_higher = sys.version_info >= (3, 11)

PROTON_VPN_SPLIT_TUNNELING = "me.proton.vpn.split_tunneling"


async def _async_proton_dbus_service_exists():
    """
    Asynchronously detects whether there is a proton dbus service available.
    Returns true if there is a proton dbus service available.
    """
    bus = await MessageBus(bus_type=BusType.SYSTEM).connect()

    reply = await bus.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="ListNames",
        )
    )

    if reply.message_type == MessageType.ERROR:
        raise RuntimeError(reply.body[0])

    names = reply.body[0]

    return PROTON_VPN_SPLIT_TUNNELING in names


def _proton_dbus_service_exists():
    """
    Synchronously detects whether there is a proton dbus service available.
    Returns true if there is a proton dbus service available.

    This runs in a separate thread which allows this function to be called
    from synchronous code running in an async context.
    """

    executor = ThreadPoolExecutor(max_workers=1)

    def _worker():
        # This runs in the new thread, where no loop is active.
        return asyncio.run(_async_proton_dbus_service_exists())

    future = executor.submit(_worker)
    return future.result()


async def _backwards_compatible_asyncio_timeout(timeout: int, func, *args):
    """
    Helper function to maintain compatibility across Python versions.

    Python 3.11 introduced `asyncio.timeout`, which is a context manager that
    can be used to set a timeout for an asynchronous operation. In earlier
    versions of Python, we used `asyncio.wait_for` to achieve similar functionality.
    """
    if is_python_3_11_or_higher:
        async with asyncio.timeout(timeout):
            await func(*args)
    else:
        await asyncio.wait_for(
            func(*args), timeout=timeout
        )


class SplitTunnelingDbusClient(SplitTunneling):
    """Split tunneling service that abstracts from necessary initializations.

    Use this class to talk to our backend daemon.
    """

    DEFAULT_TIMEOUT: int = 5

    def __init__(self, uid: int, interface: str, timeout: int = DEFAULT_TIMEOUT):
        super().__init__(uid)
        self._interface = interface
        self._timeout = timeout

    @staticmethod
    async def build(uid: int) -> SplitTunnelingDbusClient:
        """Initializes the daemon.

        Returns:
            SplitTunnelingDbusClient: new instance of the daemon
        """
        bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
        try:
            introspection = await bus.introspect(
                "me.proton.vpn.split_tunneling", "/me/proton/vpn/split_tunneling"
            )
        except dbus_fast.errors.DBusError as excp:
            raise exceptions.SplitTunnelingError(
                "Unable to start introspection"
            ) from excp
        obj = bus.get_proxy_object(
            "me.proton.vpn.split_tunneling", "/me/proton/vpn/split_tunneling", introspection
        )
        return SplitTunnelingDbusClient(
            uid=uid, interface=obj.get_interface("me.proton.vpn.split_tunneling")
        )

    async def set_config(self, config: SplitTunnelingConfig) -> None:
        """Sets a new config for instance uid.

        Raises:
            exceptions.SplitTunnelingError: whenever there is a dbus exception

        Args:
            config (SplitTunnelingConfig): the object containing the data
        """
        dbus_dict = translator.to_dbus_dict(config)
        try:
            await _backwards_compatible_asyncio_timeout(
                self._timeout,
                self._interface.call_set_config,
                self._uid, dbus_dict
            )
        except TimeoutError as exc:
            raise exceptions.SplitTunnelingError(
                "Timeout setting split tunnelint configuration"
            ) from exc
        except dbus_fast.errors.DBusError as excp:
            raise exceptions.SplitTunnelingError(
                "Error setting new split tunneling "
                f"configuration for {self._uid}"
            ) from excp
        except Exception as exc:
            raise exceptions.SplitTunnelingError(
                "Unexpected error"
            ) from exc

    async def get_config(self) -> Optional[SplitTunnelingConfig]:
        """Returns config for instance uid.

        Raises:
            exceptions.SplitTunnelingError: whenever there is a dbus exception

        Returns:
            SplitTunnelingConfig: data stored for the specified uid
        """
        try:
            dbus_dict = await self._interface.call_get_config(self._uid)
        except dbus_fast.errors.DBusError as excp:
            raise exceptions.SplitTunnelingError(
                f"Error getting split tunneling configuration for {self._uid}"
            ) from excp
        except Exception as exc:
            raise exceptions.SplitTunnelingError(
                "Unexpected error"
            ) from exc

        if not dbus_dict:
            return None

        return translator.from_dbus_dict(dbus_dict)

    async def clear_config(self) -> None:
        """Clears config stored for instance uid.

        Raises:
            exceptions.SplitTunnelingError: whenever there is a dbus exception
        """
        try:
            await _backwards_compatible_asyncio_timeout(
                self._timeout,
                self._interface.call_clear_config,
                self._uid
            )
        except TimeoutError as exc:
            raise exceptions.SplitTunnelingError(
                "Timeout clearing split tunnelint configuration"
            ) from exc
        except dbus_fast.errors.DBusError as excp:
            raise exceptions.SplitTunnelingError(
                f"Error clearing split tunneling config for {self._uid}"
            ) from excp
        except Exception as exc:
            raise exceptions.SplitTunnelingError(
                "Unexpected error"
            ) from exc

    async def get_all_configs(
            self
    ) -> Union[
        list[tuple[int, SplitTunnelingConfig]],
        list
    ]:
        """Returns a list of all configs.

        Raises:
            exceptions.SplitTunnelingError: whenever there is a dbus exception

        Returns:
            list[Optional[tuple[int, SplitTunnelingConfig]]]: \
                all stored configs
        """
        try:
            all_configs = await self._interface.call_get_all_configs()
        except dbus_fast.errors.DBusError as excp:
            raise exceptions.SplitTunnelingError(
                "Error getting all configs"
            ) from excp

        list_of_parsed_configs = []
        for uid, config in all_configs:
            list_of_parsed_configs.append(
                (uid, translator.from_dbus_dict(config))
            )

        return list_of_parsed_configs

    @classmethod
    def _get_priority(cls) -> int:
        """
        Priority of the split tunneling implementation.

        To be implemented by subclasses.
        """
        return 1

    @classmethod
    def _validate(cls) -> bool:
        """
        Determines whether the split tunneling connection
        implementation is valid or not.
        """
        service_exists = _proton_dbus_service_exists()
        if not service_exists:
            return False

        # Check if the BCC version is at least 0.26.0
        # This is necessary for the split tunneling to work correctly.
        if version.parse(bcc_version) < version.parse("0.26.0"):
            return False

        return True



#!/usr/bin/env python3
"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass

from dbus_fast.aio import MessageBus
from dbus_fast.service import ServiceInterface, method
from dbus_fast import BusType

from proton.vpn import logging
from proton.vpn.core.settings import SplitTunnelingConfig
from proton.vpn.daemon.split_tunneling import dbus_translator as translator
from proton.vpn.daemon.split_tunneling.split_tunneling import SplitTunnelingService

logger = logging.getLogger(__name__)


@dataclass
class UserConfig:
    """Associates `SplitTunnelingConfig` object to uid
    """
    uid: int
    config: SplitTunnelingConfig


class SplitTunnelingDbus(ServiceInterface):
    """Split tunneling dbus interface.

    This class defines the interface that is accessible via dbus.

    Args:
        ServiceInterface (str): The name of the dbus service.
    """

    def __init__(self):
        super().__init__("me.proton.vpn.split_tunneling")
        self._service = SplitTunnelingService()

    @method(name="SetConfig")
    async def set_config(self, uid: "q", config: "a{sv}"):  # type: ignore # noqa: F722,F821
        """Set split tunneling config

        The reason that the `Variant` datatype is used as value for the dict,
        is because the `SplitTunnelingConfig` dict contains two types of data,
        for `mode` is a string and for `app_paths` and `ip_ranges` are
        an array of strings, so `Variant` allows us to pass different
        datatypes for the value.

        To pass values to this method, the structure should look like this:
        ```
            1000,
            {
                "mode": Variant('s', "exclude"),
                "app_paths": Variant('as', ['/snap/firefox']),
                "ip_ranges": Variant('as', ['192.168.1.1'])
            }
        ```
        If you're testing via D-Feet,
        prefix variants with `GLib.`, ie:
        ```
            1000,
            {
                "mode": GLib.Variant('s', "exclude"),
                "app_paths": GLib.Variant('as', ['/snap/firefox']),
                "ip_ranges": GLib.Variant('as', ['192.168.1.1'])
            }
        ```

        Args:
            SplitTunnelingConfig: A dict object that
            follows the dbus format as follows:
                `a`: means it's an array
                `{}`: when prefixed `a` it transforms into a dict
                `s`: key is a string
                `v`: value is a `Variant`
            uid: `q` is a uint16
        """
        config: SplitTunnelingConfig = translator.from_dbus_dict(config)
        await self._service.set_config(uid, config)

    @method(name="GetConfig")
    async def get_config(self, uid: "q") -> "a{sv}":  # type: ignore # noqa: F722,F821
        """Returns configuration for specified uid.

        Since we can not return different types of data, we return
        an array with empty data in case no matching value is found.

        Returns:
            a{sv}: An array containing the information
                `a`: means it's an array
                `{}`: when prefixed `a` it transforms into a dict
                `s`: key is a string
                `v`: value is a `Variant`

            It can also return an array with empty values.
        """
        config = self._service.get_config(uid)
        return translator.to_dbus_dict(config) if config else {}

    @method(name="ClearConfig")
    async def clear_config(self, uid: "q"):  # noqa: F821
        """Clears the config for specified uid.

        Args:
            uid (uint16): uid of the user
        """
        await self._service.clear_config(uid)

    @method(name="LogStatus")
    def log_status(self):
        """Logs the service status."""
        self._service.log_status()

    @method(name="GetAllConfigs")
    async def get_all_configs(self) -> "a(qa{sv})":  # type: ignore # noqa: F722
        """Returns all stored configs

        Returns:
            list[dict[str, Variant]]: all stored configs
        """
        return [
            (uid, translator.to_dbus_dict(config))
            for uid, config in self._service.get_all_configs()
        ]


async def init_split_tunneling_daemon():
    """Main method that configures the bus.
    """
    bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
    _ = await bus.request_name('me.proton.vpn.split_tunneling')
    bus.export('/me/proton/vpn/split_tunneling', SplitTunnelingDbus())



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
from typing import Optional

from proton.vpn import logging
from proton.vpn.core.settings import SplitTunnelingConfig

from proton.vpn.daemon.split_tunneling.apps.service import AppBasedSplitTunnelingService

logger = logging.getLogger(__name__)


class SplitTunnelingService:
    """Implements the Split Tunneling feature."""

    def __init__(
            self,
            config_by_uid: Optional[dict[int, SplitTunnelingConfig]] = None,
            app_service: Optional[AppBasedSplitTunnelingService] = None
    ):
        self._config_by_uid = config_by_uid or {}
        self._app_split_tunneling = app_service or AppBasedSplitTunnelingService()
        self._lock = asyncio.Lock()

    async def set_config(self, uid: int, config: SplitTunnelingConfig):
        """
        Applies split tunneling configuration.
        @param uid: the unix user ID to apply the configuration for.
        @param config: the split tunneling configuration.
        """
        async with self._lock:
            logger.info("Setting %s for user %s", config, uid)
            self._config_by_uid[uid] = config

            if self._is_app_based_config():
                self._app_split_tunneling.start(self._config_by_uid)
            else:
                await self._app_split_tunneling.stop()

    async def clear_config(self, uid: int):
        """
        Clears split tunneling configuration for the specified user id:
        @param uid: the unix user ID to clear the configuration for.
        """
        async with self._lock:
            logger.info("Clearing config for user %s", uid)
            if uid in self._config_by_uid:
                del self._config_by_uid[uid]

                if self._is_app_based_config():
                    self._app_split_tunneling.start(self._config_by_uid)
                else:
                    await self._app_split_tunneling.stop()

    def get_config(self, uid: int) -> Optional[SplitTunnelingConfig]:
        """
        Returns the split tunneling configuration for a given user id.
        @param uid: the unix user id.
        @returns: the split tunneling configuration, if existing. Otherwise `None`.
        """
        return self._config_by_uid.get(uid)

    def get_all_configs(self) -> list[tuple[int, SplitTunnelingConfig]]:
        """
        Returns all available split tunneling configurations.
        @returns: a list of tuples where each tuple contains the user id
            and the split tunneling configuration, in this order.
        """
        return list(self._config_by_uid.items())

    def log_status(self):
        """Log ST service status"""
        logger.info("============Split Tunneling service status===========")
        self._app_split_tunneling.log_status()
        logger.info("=====================================================")

    def _is_app_based_config(self) -> bool:
        return any(
            any(path for path in config.app_paths)  # any non empty path
            for config in self._config_by_uid.values()    # on any user config
        )



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass, field
import time

import psutil

from proton.vpn.core.settings import SplitTunnelingConfig, SplitTunnelingMode
from proton.vpn import logging

logger = logging.getLogger(__name__)


@dataclass
class Process:
    """
    Holds all required process information to do Split Tunneling.
    """
    uid: int
    pid: int
    ppid: int
    exe: str

    # configured ST paths this process matched against
    matched_config_paths: set[str] = field(default_factory=set)

    # whether the process is still running or it already exited
    running: bool = True

    @staticmethod
    def from_psutil(process: psutil.Process):
        """
        Builds a Process instance from a psutil.Process instance.
        """
        uid = process.uids().real
        ppid = process.ppid()
        exe = ""
        try:
            exe = process.exe()
        except (psutil.AccessDenied, psutil.ZombieProcess) as error:
            # Even when running as root, some processes raise this
            logger.warning("Error getting process path: %s: %s", type(error).__name__, error)

        return Process(
            pid=process.pid, uid=uid, ppid=ppid, exe=exe
        )


class ProcessMatcher:
    """Matches processes against ST configuration."""

    VPN_APP_BIN = "protonvpn-app"

    @classmethod
    def check_process(
            cls, process: Process, config_by_uid: dict[int, SplitTunnelingConfig]
    ) -> set[str]:
        """
        Given a process and the ST config, it returns a set with all the app paths in the
        ST config the process matched against.
        :param process: process to check for matches
        :param config_by_uid: ST config indexed by user ID (unix UID).
        :returns: the app path matches.
        """
        if process.uid not in config_by_uid:
            return set()

        # In include mode, the Proton VPN app should always be included in the VPN
        # so that it can establish the local agent connection to the VPN server.
        config = config_by_uid[process.uid]
        if config.mode == SplitTunnelingMode.INCLUDE and cls.VPN_APP_BIN in process.exe:
            return set([cls.VPN_APP_BIN])

        matches = set()
        for app_path in config.app_paths:
            if not app_path:
                continue
            if process.exe.startswith(app_path):
                matches.add(app_path)

        return matches

    @classmethod
    def check_all_processes(cls, config_by_uid) -> dict[int, Process]:
        """
        Returns a dict with all currently running processes indexed by ID. Each process
        contains the configured ST app paths it matched against.

        Note: processes started by users for which there is not ST config are ignored.
        """
        start = time.time_ns()
        tracked_procs = {}

        for psutil_process in psutil.process_iter():
            process = tracked_procs.get(psutil_process.pid) or Process.from_psutil(psutil_process)
            if process.uid not in config_by_uid:
                # only processes created by users that added ST config are checked
                continue

            if process.matched_config_paths:
                # already tracked child of a process matching config paths.
                continue

            matched_config_paths = cls.check_process(process, config_by_uid)
            process.matched_config_paths.update(matched_config_paths)
            tracked_procs[process.pid] = process

            if process.matched_config_paths:
                # if the process matched any config paths, then we also flag all its
                # descendant processes with the same config path matches
                for descendant in psutil_process.children(recursive=True):
                    descendant = tracked_procs.get(descendant.pid) \
                        or Process.from_psutil(descendant)
                    descendant.matched_config_paths.update(matched_config_paths)
                    tracked_procs[descendant.pid] = descendant

        logger.info("Existing processes inspected in %d ms", (time.time_ns() - start) // 1_000_000)

        return tracked_procs



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from importlib.metadata import version
import asyncio

from proton.vpn import logging
from proton.vpn.daemon.split_tunneling.service import \
    init_split_tunneling_daemon
from proton.vpn.daemon.exception_handler import asyncio_exception_handler


async def main():
    """Main method to call daemons.
    """

    await init_split_tunneling_daemon()


def run_forever():
    """Runs the loop forever
    """
    logging.config(filename="vpn-daemon")
    logger = logging.getLogger(__name__)

    logger.info("Starting Proton VPN daemon v%s", version("proton-vpn-daemon"))
    loop = asyncio.new_event_loop()

    # Configure the exception handler for asyncio.
    loop.set_exception_handler(asyncio_exception_handler)

    loop.run_until_complete(main())
    loop.run_forever()

    # Restore to original exception handler.
    loop.set_exception_handler(None)


if __name__ == '__main__':
    try:
        run_forever()
    except KeyboardInterrupt:
        print("Service stopped.")



#!/usr/bin/env python3
"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Callable, Optional

import asyncio

from bcc import BPF

from proton.vpn import logging
from proton.vpn.core.settings import SplitTunnelingConfig, SplitTunnelingMode

from proton.vpn.daemon.split_tunneling.apps.process_map import ProcessMap
from proton.vpn.daemon.split_tunneling.apps.process_matcher import \
    Process, ProcessMatcher

logger = logging.getLogger(__name__)

BPF_PROGRAM_PATH = Path(__file__.replace(".py", ".bpf.c"))


class PerfBufferEventType(Enum):
    """Type of events sent to the BPF perf buffer"""
    EXEC_ARGV_FRAGMENT = 0
    EXEC = 1
    CLONE = 2
    EXIT = 3


# pylint: disable=too-many-instance-attributes,
class ProcessMonitor:
    """Monitors processes based on the specified Split Tunneling configuration."""

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            process_event_callback: Callable[[Process, SplitTunnelingMode]],
            config_by_uid: Optional[dict[int, SplitTunnelingConfig]] = None,
            tracked_procs: Optional[dict[int, Process]] = None,
            bpf: Optional[BPF] = None,
            process_matcher: Optional[ProcessMatcher] = None
    ):
        self._process_event_callback = process_event_callback
        self._config_by_uid = config_by_uid
        self._bpf = bpf
        self._process_matcher = process_matcher or ProcessMatcher()
        self._background_task: Optional[asyncio.Task] = None
        self._stop_requested = False
        self._tracked_procs = tracked_procs or {}
        self._argv = defaultdict(list)

    @property
    def config_by_uid(self) -> Optional[dict[int, SplitTunnelingConfig]]:
        """Returns the curret config in use."""
        return self._config_by_uid

    def log_status(self):
        """Logs the process monitor status."""
        logger.info("===============Process monitor status================")
        logger.info("Config: %s", self._config_by_uid)
        process_map = ProcessMap(self._tracked_procs)
        logger.info("Tracked processes dumped at: %s", process_map.dump())
        logger.info("Tracked process trees dumped at: %s", process_map.dump_process_trees())
        logger.info("=====================================================")

    def start(
            self, config_by_uid: dict[int, SplitTunnelingConfig]
    ) -> asyncio.Task:
        """
        Starts a background task to monitor processes.
        @param config_by_uid: map of split tunneling configuration by uid (unix user id)
        @param process_event_callback: callback called whenever there is a process event.

        Note that this method returns straight away, it doesn't wait that
        the background task is running.
        """
        self._config_by_uid = config_by_uid

        if not self._background_task:
            logger.info("Starting process monitor")

            self._tracked_procs = self._process_matcher.check_all_processes(config_by_uid)
            for process in self._tracked_procs.values():
                self._process_event_callback(process, config_by_uid[process.uid].mode)

            self._attach_bpf()

            # start listening for process events via ebpf
            self._background_task = asyncio.create_task(
                self._run_process_monitoring()
            )

            # Ensure exceptions are bubbled up and caught by the exception handler
            def on_done(future):
                try:
                    future.result()
                except asyncio.CancelledError:
                    pass

            self._background_task.add_done_callback(on_done)
        else:
            logger.info("Process monitor already running: config updated")

        return self._background_task

    async def stop(self):
        """
        Triggers a shutdown request to the background task monitoring processes,
        and waits for it to finish.
        """
        if not self._background_task:
            logger.info("Process monitor is already stopped.")
            return

        logger.info("Stopping process monitor")
        self._stop_requested = True
        self._detach_bpf()
        try:
            await self._background_task
        except Exception:  # pylint: disable=broad-except
            logger.exception("Unexpected error while monitoring processes")
        # reset object state
        callback = self._process_event_callback
        self.__init__(process_event_callback=callback)  # pylint: disable=C2801
        logger.info("Process monitor stopped")

    def _attach_bpf(self):
        if not self._bpf:
            with open(BPF_PROGRAM_PATH, "r", encoding="utf-8") as file:
                bpf_text = file.read()
            try:
                self._bpf = BPF(text=bpf_text)
            except Exception as exc:  # pylint: disable=broad-except
                if f"{exc}".startswith("Failed to compile"):
                    raise RuntimeError(
                        "Failed to start process monitor. Kernel headers might be missing "
                        "or not compatible with the current kernel version."
                    ) from exc
                raise
            # All probes use TRACEPOINT_PROBE and are automatically
            # attached when the BPF object is created.

    def _detach_bpf(self):
        # All probes use TRACEPOINT_PROBE and are automatically
        # detached when the BPF object is destroyed.
        self._bpf = None

    async def _run_process_monitoring(self):
        await asyncio.get_running_loop().run_in_executor(
            None, self._run_blocking_process_monitoring
        )

    def _run_blocking_process_monitoring(self):
        self._bpf["events"].open_perf_buffer(self._process_perf_buffer_event)
        while not self._stop_requested:
            self._bpf.perf_buffer_poll(timeout=30)  # timeout in ms

    def _process_perf_buffer_event(self, _cpu, data, _size):
        event = self._bpf["events"].event(data)

        if event.type == PerfBufferEventType.EXEC_ARGV_FRAGMENT.value:
            # Whenever an exec syscall is initiated, multiple EXEC_ARGV_FRAGMENT events are
            # sent, one per argv: the first one contains the full path, and the following
            # ones contain each argument (up to a limit, see ebpf program).
            self._argv[event.pid].append(event.argv)
            return

        exe = ""
        if event.type == PerfBufferEventType.EXEC.value:
            # Once the exec syscall returns, the final exe path is built out of all the
            # previous EXEC_ARGV_FRAGMENT events containing the fragments that make it up.
            exe = b' '.join(self._argv[event.pid]).replace(b'\n', b'\\n').decode('utf-8')
            try:
                del self._argv[event.pid]
            except KeyError:
                pass

        self.check_for_config_matches(
            PerfBufferEventType(value=event.type),
            Process(event.uid, event.pid, event.ppid, exe)
        )

    def check_for_config_matches(self, event: PerfBufferEventType, process: Process):
        """Checks if the process matches the ST config and calls the callback."""
        if process.uid not in self.config_by_uid:
            # processes started by users that didn't set any ST config are ignored
            return

        if event is PerfBufferEventType.EXEC:
            already_tracked_process = self._tracked_procs.get(process.pid)
            if already_tracked_process and already_tracked_process.matched_config_paths:
                # We do a sticky process matching: once a process matches one of the
                # config paths then any subsequent exec syscalls done by the same process
                # are ignored. This is because some apps start from an executable
                # but then run an exec syscall to a different one (e.g. Google Chrome).
                return

            # a new process was started: check for config path matches
            matches = self._process_matcher.check_process(process, self._config_by_uid)
            process.matched_config_paths.update(matches)
            self._tracked_procs[process.pid] = process
            self._process_event_callback(process, self.config_by_uid[process.uid].mode)

        elif event is PerfBufferEventType.CLONE:
            # a process was cloned/forked: check if its parent matched config paths
            parent = self._tracked_procs.get(process.ppid)
            if parent:
                # a cloned process inherits its parent config path matches
                process.matched_config_paths.update(parent.matched_config_paths)
                self._tracked_procs[process.pid] = process
                self._process_event_callback(process, self.config_by_uid[process.uid].mode)

        elif event is PerfBufferEventType.EXIT:
            # a process exited: stop tracking it
            if process.pid in self._tracked_procs:
                process = self._tracked_procs[process.pid]
                process.running = False
                self._process_event_callback(process, self.config_by_uid[process.uid].mode)
                del self._tracked_procs[process.pid]


def build_process_monitor_cli_parser(name: str):
    """Builds the process monitor CLI arg parser"""
    import argparse  # pylint: disable=C0415

    parser = argparse.ArgumentParser(
        prog=name,
        description="For testing purposes only"
    )
    parser.add_argument("-u", "--uid", required=True, type=int, help="UID to exclude process for.")
    parser.add_argument(
        "-m", "--mode", required=True, choices=["exclude", "include"],
        help="Split Tunneling mode"
    )
    parser.add_argument(
        "-p", "--path", required=True, action="append",
        help="Process paths to exclude."
    )

    return parser


async def main():
    """Test script"""
    logging.config(filename="stprocmon")
    _logger = logging.getLogger("stprocmon")

    parser = build_process_monitor_cli_parser(
        name="Process monitor for app-based Split Tunneling"
    )
    args = parser.parse_args()

    def callback(process: Process, mode: SplitTunnelingMode):  # pylint: disable=unused-argument
        if process.matched_config_paths:
            for config_path in args.path:
                if any(
                    process_path.startswith(config_path)
                    for process_path in process.matched_config_paths
                ):
                    _logger.info("Process match: %s", process)

    process_monitor = ProcessMonitor(process_event_callback=callback)
    try:
        await process_monitor.start(
            config_by_uid={
                args.uid: SplitTunnelingConfig(
                    mode=SplitTunnelingMode(value=args.mode),
                    app_paths=args.path
                )
            }
        )
    except asyncio.CancelledError:
        pass
    finally:
        try:
            await process_monitor.stop()
        except asyncio.CancelledError:
            pass


if __name__ == "__main__":
    asyncio.run(main())



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Optional
import ctypes
import os
import re
import subprocess  # nosec # nosemgrep: gitlab.bandit.B404

from bcc import BPF

from proton.vpn import logging

logger = logging.getLogger(__name__)


_FWMARK_MAP_KEY = ctypes.c_int32(1)


# ebpf program to split traffic based on PID map
BPF_PROGRAM = """
BPF_HASH(fwmark_map, u32, u32);
BPF_HASH(pid_map, u32, u32);

int split_tunnel(struct bpf_sock *sk) {
    u32 pid = bpf_get_current_pid_tgid();

    u32 fwmark_key = 1;
    u32 *fwmark = fwmark_map.lookup(&fwmark_key);

    u32 *pid_found = pid_map.lookup(&pid);

    if (fwmark && pid_found) {
        bpf_trace_printk("Excluding PID %d from VPN\\n", pid);
        sk->mark = *fwmark;
    }

    return 1;
}
"""


def _get_wireguard_fwmark(interface) -> Optional[int]:
    # FIXME: get fwmark via D-Bus from NetworkManager  # pylint: disable=fixme
    #  instead of with sudo wg show. The WG backend could somehow make it available?
    try:
        result = subprocess.run(  # nosec
            ["/usr/bin/wg", "show", interface], capture_output=True, text=True, check=True
        )
        fwmark_match = re.search(r'fwmark:\s+(\S+)', result.stdout)
        if fwmark_match:
            fwmark_hex = fwmark_match.group(1)
            fwmark_int = int(fwmark_hex, 16)
            return fwmark_int

        raise RuntimeError(
            f"Couldn't get fwmark from interface {interface}: \n\n{result.stdout}"
        )
    except subprocess.CalledProcessError:
        # `wg show proton0` failed: the user is not connected to the VPN
        return None


class SocketMonitor:
    """Split-tunnels sockets based on the proces that creates them."""

    # FIXME could the WG backend somehow pass the interface name (proton0)  # pylint: disable=fixme
    WIREGUARD_INTERFACE_NAME = "proton0"

    def __init__(self):
        self._bpf = BPF(text=BPF_PROGRAM)
        self._bpf_fwmark_map = self._bpf.get_table("fwmark_map")
        self._bpf_pid_map = self._bpf.get_table("pid_map")
        self._bpf_split_tunneling_func = self._bpf.load_func(
            "split_tunnel", self._bpf.CGROUP_SOCK
        )
        self._cgroup = None
        self._bpf_enum_group = None

    def log_status(self):
        """Logs the socket monitor status."""
        logger.info("==============Socket monitor status==================")
        logger.info("Tracked PIDs: %s", [key.value for key in self._bpf_pid_map.keys()])
        logger.info("fwmark: %s", self._bpf_fwmark_map.get(_FWMARK_MAP_KEY))
        logger.info("=====================================================")

    def start(self):
        """Starts monitoring sockets."""
        fwmark = _get_wireguard_fwmark(self.WIREGUARD_INTERFACE_NAME)
        if fwmark is not None:
            self._bpf_fwmark_map[_FWMARK_MAP_KEY] = ctypes.c_int32(fwmark)
        elif _FWMARK_MAP_KEY in self._bpf_fwmark_map:
            del self._bpf_fwmark_map[_FWMARK_MAP_KEY]

        if self._started:
            logger.info("Socket monitor already running: fwmark updated to %s", fwmark)
            return

        logger.info("Starting socket monitor (with fwmark %s)", fwmark)
        self._cgroup = os.open("/sys/fs/cgroup/user.slice", os.O_RDONLY)

        # attach the ST eBPF function to the cgroup/sock_create event
        # of the root user cgroup
        self._bpf.attach_func(
            self._bpf_split_tunneling_func,
            self._cgroup,
            self._backwards_compatible_bfp_attach_type.CGROUP_INET_SOCK_CREATE
        )

    @property
    def _started(self):
        return self._cgroup is not None

    def _cleanup(self):
        self._bpf_pid_map.clear()
        self._bpf_fwmark_map.clear()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop()

    def exclude_process_from_vpn(self, pid: int):  # pylint: disable=redefined-outer-name
        """
        Split tunnels the specified process. Afterwards, traffic generated by sockets
        created by such process will be redirected outside the VPN.
        @param pid: the unix id (pid) of the process to split tunnel.
        """
        if not self._started:
            raise RuntimeError("Socket monitor was not started yet")

        logger.debug("Adding %s to pid map", pid)
        self._bpf_pid_map[ctypes.c_uint32(pid)] = ctypes.c_uint32(1)

    def stop_tracking_process(self, pid: int):  # pylint: disable=redefined-outer-name
        """
        Stops tracking sockets opened by the specified process process.
        @param pid: the unix id of the process.
        If the specified process was not being split tunneled then this is a noop.
        """
        if not self._started:
            raise RuntimeError("Socket monitor was not started yet")

        pid_c_uint32 = ctypes.c_uint32(pid)
        if pid_c_uint32 in self._bpf_pid_map:
            logger.debug("Removing %s from pid map", pid)
            del self._bpf_pid_map[pid_c_uint32]

    def stop(self):
        """Stops monitoring sockets."""
        if not self._started:
            logger.info("Socket monitor is already stopped")
            return

        logger.info("Unloading split tunneling ebpf function")
        self._bpf.detach_func(
            self._bpf_split_tunneling_func,
            self._cgroup,
            self._backwards_compatible_bfp_attach_type.CGROUP_INET_SOCK_CREATE
        )
        self._cleanup()
        os.close(self._cgroup)
        self._cgroup = None

        logger.info("Socket monitor stopped")

    @property
    def _backwards_compatible_bfp_attach_type(self) -> object:
        """In v20 of bcc a refactor was made where enums
        were extracted into their own type, the BPFAttachType.

        Before that the types were part of the bpf program.

        See more here:
        https://github.com/iovisor/bcc/commit/2731825b9327a9a720f2ef92ed891ce0525a8dc3

        Returns:
            object: Either the BPFAttachType or bpf program.
        """
        if not self._bpf_enum_group:
            try:
                from bcc import BPFAttachType  # pylint: disable=import-outside-toplevel
                self._bpf_enum_group = BPFAttachType
            except ImportError:
                self._bpf_enum_group = self._bpf

        return self._bpf_enum_group


def main():
    """Test script"""
    is_child_process = os.fork()
    if is_child_process:
        pid = os.getpid()
        with SocketMonitor() as socket_monitor:
            socket_monitor.exclude_process_from_vpn(pid)
            from urllib.request import Request, urlopen  # pylint: disable=import-outside-toplevel
            request = Request("https://ip.me", headers={"User-Agent": "curl/7.54.1"})
            # nosemgrep: python.lang.security.audit.dynamic-urllib-use-detected.dynamic-urllib-use-detected # pylint: disable=line-too-long # noqa: E501
            with urlopen(request) as response:  # nosemgrep: gitlab.bandit.B310-1 # nosec B310
                current_ip = response.read().decode('utf-8')
                print(f"Current IP: {current_ip}")
